@@ -26,7 +26,6 @@
 *	8/18/93 - Mark Edel & Joy Kyriakopulos - Ported to VMS		       *
 *									       *
 *******************************************************************************/
-static char SCCSID[] = "@(#)nedit.c	1.27     10/31/94";
 #include <stdio.h>
 #include <stdlib.h>
 #ifdef USE_XMIM
@@ -54,13 +53,17 @@ static char SCCSID[] = "@(#)nedit.c	1.27     10/31/94";
 #include "nedit.h"
 #include "file.h"
 #include "preferences.h"
+#include "regularExp.h"
 #include "selection.h"
 #include "tags.h"
 #include "menu.h"
 #include "macro.h"
 #include "server.h"
+#include "interpret.h"
+#include "parse.h"
 
 static void nextArg(int argc, char **argv, int *argIndex);
+static int checkDoMacroArg(char *macro);
 
 WindowInfo *WindowList = NULL;
 Display *TheDisplay;
@@ -83,6 +86,10 @@ static char *fallbackResources[] = {
     "*text.highlightForeground: black",
     "*XmText*foreground: black",
     "*XmText*background: #cccccc",
+    "*helpText.background: #cccccc",
+    "*helpText.foreground: black",
+    "*helpText.selectBackground: #b3b3b3",
+    "*helpText.font: -adobe-courier-medium-r-normal-*-14-*-*-*-*-*-*-*",
     "*XmText.translations: #override \
 Ctrl~Alt~Meta<KeyPress>v: paste-clipboard()\n\
 Ctrl~Alt~Meta<KeyPress>c: copy-clipboard()\n\
@@ -129,7 +136,7 @@ Ctrl~Alt~Meta<KeyPress>x: cut-clipboard()\n"
     "*editMenu.paste.acceleratorText: Ctrl+V",
     "*editMenu.pasteColumn.accelerator: Shift Ctrl<Key>v",
     "*editMenu.pasteColumn.acceleratorText: Ctrl+Shift+V",
-    "*editMenu.clear.acceleratorText: Del",
+    "*editMenu.delete.acceleratorText: Del",
     "*editMenu.selectAll.accelerator: Ctrl<Key>a",
     "*editMenu.selectAll.acceleratorText: Ctrl+A",
     "*editMenu.shiftLeft.accelerator: Ctrl<Key>9",
@@ -138,8 +145,8 @@ Ctrl~Alt~Meta<KeyPress>x: cut-clipboard()\n"
     "*editMenu.shiftRight.accelerator: Ctrl<Key>0",
     "*editMenu.shiftRight.acceleratorText: [Shift]Ctrl+0",
     "*editMenu.shiftRightShift.accelerator: Shift Ctrl<Key>0",
-    "*editMenu.capitalize.accelerator: Ctrl<Key>6",
-    "*editMenu.capitalize.acceleratorText: Ctrl+6",
+    "*editMenu.upperCase.accelerator: Ctrl<Key>6",
+    "*editMenu.upperCase.acceleratorText: Ctrl+6",
     "*editMenu.lowerCase.accelerator: Shift Ctrl<Key>6",
     "*editMenu.lowerCase.acceleratorText: Shift+Ctrl+6",
     "*editMenu.fillParagraph.accelerator: Ctrl<Key>j",
@@ -152,18 +159,18 @@ Ctrl~Alt~Meta<KeyPress>x: cut-clipboard()\n"
     "*searchMenu.find.accelerator: Ctrl<Key>f",
     "*searchMenu.find.acceleratorText: [Shift]Ctrl+F",
     "*searchMenu.findShift.accelerator: Shift Ctrl<Key>f",
-    "*searchMenu.findSame.accelerator: Ctrl<Key>g",
-    "*searchMenu.findSame.acceleratorText: [Shift]Ctrl+G",
-    "*searchMenu.findSameShift.accelerator: Shift Ctrl<Key>g",
+    "*searchMenu.findAgain.accelerator: Ctrl<Key>g",
+    "*searchMenu.findAgain.acceleratorText: [Shift]Ctrl+G",
+    "*searchMenu.findAgainShift.accelerator: Shift Ctrl<Key>g",
     "*searchMenu.findSelection.accelerator: Ctrl<Key>h",
     "*searchMenu.findSelection.acceleratorText: [Shift]Ctrl+H",
     "*searchMenu.findSelectionShift.accelerator: Shift Ctrl<Key>h",
     "*searchMenu.replace.accelerator: Ctrl<Key>r",
     "*searchMenu.replace.acceleratorText: [Shift]Ctrl+R",
     "*searchMenu.replaceShift.accelerator: Shift Ctrl<Key>r",
-    "*searchMenu.replaceSame.accelerator: Ctrl<Key>t",
-    "*searchMenu.replaceSame.acceleratorText: [Shift]Ctrl+T",
-    "*searchMenu.replaceSameShift.accelerator: Shift Ctrl<Key>t",
+    "*searchMenu.replaceAgain.accelerator: Ctrl<Key>t",
+    "*searchMenu.replaceAgain.acceleratorText: [Shift]Ctrl+T",
+    "*searchMenu.replaceAgainShift.accelerator: Shift Ctrl<Key>t",
     "*searchMenu.gotoLineNumber.accelerator: Ctrl<Key>l",
     "*searchMenu.gotoLineNumber.acceleratorText: Ctrl+L",
     "*searchMenu.gotoSelected.accelerator: Ctrl<Key>e",
@@ -171,16 +178,17 @@ Ctrl~Alt~Meta<KeyPress>x: cut-clipboard()\n"
     "*searchMenu.mark.accelerator: Alt<Key>m",
     "*searchMenu.mark.acceleratorText: Alt+M a-z",
     "*searchMenu.gotoMark.accelerator: Alt<Key>g",
-    "*searchMenu.gotoMark.acceleratorText: Alt+G a-z",
+    "*searchMenu.gotoMark.acceleratorText: [Shift]Alt+G a-z",
+    "*searchMenu.gotoMarkShift.accelerator: Shift Alt<Key>g",
     "*searchMenu.match.accelerator: Ctrl<Key>m",
     "*searchMenu.match.acceleratorText: Ctrl+M",
     "*searchMenu.findDefinition.accelerator: Ctrl<Key>d",
     "*searchMenu.findDefinition.acceleratorText: Ctrl+D",
     "*preferencesMenu.mnemonic: P",
-    "*preferencesMenu.overstrike.accelerator: Ctrl<Key>b",
-    "*preferencesMenu.overstrike.acceleratorText: Ctrl+B",
     "*preferencesMenu.statisticsLine.accelerator: Alt<Key>a",
     "*preferencesMenu.statisticsLine.acceleratorText: Alt+A",
+    "*preferencesMenu.overtype.accelerator: Ctrl<Key>b",
+    "*preferencesMenu.overtype.acceleratorText: Ctrl+B",
     "*shellMenu.mnemonic: l",
     "*shellMenu.filterSelection.accelerator: Alt<Key>r",
     "*shellMenu.filterSelection.acceleratorText: Alt+R",
@@ -199,6 +207,8 @@ Ctrl~Alt~Meta<KeyPress>x: cut-clipboard()\n"
     "*macroMenu.cancelLearn.acceleratorText: Ctrl+.",
     "*macroMenu.replayKeystrokes.accelerator: Ctrl<Key>k",
     "*macroMenu.replayKeystrokes.acceleratorText: Ctrl+K",
+    "*macroMenu.repeat.accelerator: Ctrl<Key>comma",
+    "*macroMenu.repeat.acceleratorText: Ctrl+,",
     "*windowsMenu.mnemonic: W",
     "*windowsMenu.splitWindow.accelerator: Ctrl<Key>2",
     "*windowsMenu.splitWindow.acceleratorText: Ctrl+2",
@@ -215,7 +225,7 @@ static char cmdLineHelp[] =
 	       [-noautoindent] [-autosave] [-noautosave] [-rows n]\n\
 	       [-columns n] [-font font] [-display [host]:server[.screen]\n\
 	       [-geometry geometry] [-xrm resourcestring] [-svrname name]\n\
-	       [file...]\n";
+	       [-import file] [file...]\n";
 #else
 "";
 #endif /*VMS*/
@@ -223,10 +233,10 @@ static char cmdLineHelp[] =
 int main(int argc, char **argv)
 {
     int i, lineNum, nRead, fileSpecified = FALSE, editFlags = CREATE;
-    int isServer = FALSE, gotoLine = False;
-    char *errMsg, *toDoCommand = NULL, *doPtr;
+    int isServer = FALSE, gotoLine = False, macroFileRead = False;
+    char *stoppedAt, *errMsg, *toDoCommand = NULL;
     char filename[MAXPATHLEN], pathname[MAXPATHLEN];
-    actionList *actionL;
+    Program *prog;
     XtAppContext context;
     XrmDatabase prefDB;
 
@@ -251,7 +261,9 @@ int main(int argc, char **argv)
         fprintf(stderr,"NEdit: cannot set locale modifiers.\n");
 #else
     /* Set locale for C library functions */
+#ifndef apollo
     setlocale(LC_CTYPE, "");
+#endif
 #endif
     
     /* Initialize toolkit and open display. */
@@ -284,6 +296,10 @@ int main(int argc, char **argv)
 	XtWarning ("NEdit: Can't open display\n");
 	exit(0);
     }
+    
+    /* Initialize global symbols and subroutines used in the macro language */
+    InitMacroGlobals();
+    RegisterMacroSubroutines();
 
     /* Store preferences from the command line and .nedit file, 
        and set the appropriate preferences */
@@ -296,10 +312,23 @@ int main(int argc, char **argv)
     /* Set up action procedures for menu item commands */
     InstallMenuActions(context);
     
+    /* Install word delimiters for regular expression matching */
+    SetREDefaultWordDelimiters(GetPrefDelimiters());
+    
     /* Read the nedit dynamic database of files for the Open Previous
        command (and eventually other information as well) */
     ReadNEditDB();
     
+    /* Process -import command line argument before others which might
+       open windows (loading preferences doesn't update menu settings,
+       which would then be out of sync with the real preference settings) */
+    for (i=1; i<argc; i++) {
+    	if (!strcmp(argv[i], "-import")) {
+    	    nextArg(argc, argv, &i);
+    	    ImportPrefFile(argv[i]);
+	}
+    }
+
     /* Process any command line arguments (-tags, -do, -read, -create,
        +<line_number>, -line, -server, and files to edit) not already
        processed by RestoreNEditPrefs. */
@@ -311,15 +340,8 @@ int main(int argc, char **argv)
     	    	fprintf(stderr, "NEdit: Unable to load tags file\n");
     	} else if (!strcmp(argv[i], "-do")) {
     	    nextArg(argc, argv, &i);
-    	    doPtr = argv[i];
-    	    actionL = ParseActionList(&doPtr, &errMsg);
-    	    if (actionL == NULL)
-    	    	fprintf(stderr, "NEdit: Unable to parse argument to do: %s\n",
-    	    		errMsg);
-    	    else {
-    	    	FreeActionList(actionL);
+	    if (checkDoMacroArg(argv[i]))
     	    	toDoCommand = argv[i];
-    	    }
     	} else if (!strcmp(argv[i], "-read")) {
     	    editFlags |= FORCE_READ_ONLY;
     	} else if (!strcmp(argv[i], "-create")) {
@@ -339,6 +361,8 @@ int main(int argc, char **argv)
     	    	gotoLine = True;
     	} else if (!strcmp(argv[i], "-server")) {
     	    isServer = True;
+	} else if (!strcmp(argv[i], "-import")) {
+	    nextArg(argc, argv, &i); /* already processed, skip */
     	} else if (*argv[i] == '-') {
 #ifdef VMS
 	    *argv[i] = '/';
@@ -357,8 +381,12 @@ int main(int argc, char **argv)
 	    for (j = 0; j < numFiles; ++j) {
 	    	ParseFilename(nameList[j], filename, pathname);
 		EditExistingFile(WindowList, filename, pathname, editFlags);
+		if (!macroFileRead) {
+		    ReadMacroInitFile(WindowList);
+		    macroFileRead = True;
+		}
 		if (toDoCommand != NULL)
-	    	    DoMacro(WindowList, toDoCommand);
+	    	    DoMacro(WindowList, toDoCommand, "-do macro");
 	    	if (gotoLine)
 	    	    SelectNumberedLine(WindowList, lineNum);
 		fileSpecified = TRUE;
@@ -370,11 +398,16 @@ int main(int argc, char **argv)
 	    ParseFilename(argv[i], filename, pathname);
 	    EditExistingFile(WindowList, filename, pathname, editFlags);
 	    fileSpecified = TRUE;
+	    if (!macroFileRead) {
+		ReadMacroInitFile(WindowList);
+		macroFileRead = True;
+	    }
 	    if (toDoCommand != NULL)
-	    	DoMacro(WindowList, toDoCommand);
+	    	DoMacro(WindowList, toDoCommand, "-do macro");
 	    if (gotoLine)
 	    	SelectNumberedLine(WindowList, lineNum);
 #endif /*VMS*/
+	    toDoCommand = NULL;
 	}
     }
 #ifdef VMS
@@ -391,10 +424,14 @@ int main(int argc, char **argv)
     /* If no file to edit was specified, open a window to edit "Untitled" */
     if (!fileSpecified) {
     	EditNewFile();
+	ReadMacroInitFile(WindowList);
 	if (toDoCommand != NULL)
-	    DoMacro(WindowList, toDoCommand);
+	    DoMacro(WindowList, toDoCommand, "-do macro");
     }
-        
+    
+    /* Begin remembering last command invoked for "Repeat" menu item */
+    AddLastCommandActionHook(context);
+
     /* Set up communication port and write ~/.nedit_server_process file */
     if (isServer)
     	InitServerCommunication();
@@ -421,3 +458,30 @@ static void nextArg(int argc, char **argv, int *argIndex)
     }
     (*argIndex)++;
 }
+
+/*
+** Return True if -do macro is valid, otherwise write an error on stderr
+*/
+static int checkDoMacroArg(char *macro)
+{
+    Program *prog;
+    char *errMsg, *stoppedAt, *tMacro;
+    int macroLen;
+    
+    /* Add a terminating newline (which command line users are likely to omit
+       since they are typically invoking a single routine) */
+    macroLen = strlen(macro);
+    tMacro = XtMalloc(strlen(macro+2));
+    strncpy(tMacro, macro, macroLen);
+    tMacro[macroLen] = '\n';
+    tMacro[macroLen+1] = '\0';
+    
+    /* Do a test parse */
+    prog = ParseMacro(tMacro, &errMsg, &stoppedAt);
+    if (prog == NULL) {
+    	ParseError(NULL, tMacro, stoppedAt, "argument to -do", errMsg);
+	return False;
+    }
+    FreeProgram(prog);
+    return True;
+}
