@@ -22,7 +22,6 @@
 * Written by Mark Edel							       *
 *									       *
 *******************************************************************************/
-static char SCCSID[] = "@(#)window.c	1.34     9/22/94";
 #include <stdio.h>
 #include <stdlib.h>
 #ifdef VMS
@@ -38,12 +37,12 @@ static char SCCSID[] = "@(#)window.c	1.34     9/22/94";
 #include <Xm/PanedW.h>
 #include <Xm/PanedWP.h>
 #include <Xm/RowColumnP.h>
-/* #include <Xm/DrawingA.h> */
 #include <Xm/Text.h>
 #include <Xm/ToggleB.h>
 #include <Xm/Label.h>
 #include <Xm/ScrolledW.h>
 #include <Xm/ScrollBar.h>
+#include <Xm/PrimitiveP.h>
 #ifdef EDITRES
 #include <X11/Xmu/Editres.h>
 extern void _XEditResCheckMessages();
@@ -53,6 +52,8 @@ extern void _XEditResCheckMessages();
 #include "textBuf.h"
 #include "textSel.h"
 #include "text.h"
+#include "textDisp.h"
+#include "textP.h"
 #include "nedit.h"
 #include "window.h"
 #include "menu.h"
@@ -61,14 +62,20 @@ extern void _XEditResCheckMessages();
 #include "undo.h"
 #include "preferences.h"
 #include "selection.h"
+#include "shell.h"
+#include "macro.h"
+#include "highlight.h"
+#include "smartIndent.h"
+#include "userCmds.h"
 #include "nedit.bm"
+#include "n.bm"
 
 /* Initial minimum height of a pane.  Just a fallback in case setPaneMinHeight
    (which may break in a future release) is not available */
 #define PANE_MIN_HEIGHT 39
 
 static Widget createTextArea(Widget parent, WindowInfo *window, int rows,
-	int cols);
+	int cols, int emTabDist, char *delimiters, int wrapMargin);
 static void addToWindowList(WindowInfo *window);
 static void removeFromWindowList(WindowInfo *window);
 static void focusCB(Widget w, WindowInfo *window, XtPointer callData);
@@ -78,10 +85,9 @@ static void movedCB(Widget w, WindowInfo *window, XtPointer callData);
 static void dragStartCB(Widget w, WindowInfo *window, XtPointer callData);
 static void dragEndCB(Widget w, WindowInfo *window, dragEndCBStruct *callData);
 static void closeCB(Widget w, WindowInfo *window, XtPointer callData);
-static void recreateWindow(WindowInfo *window, char *fontName, XmFontList fontList);
 static void setPaneDesiredHeight(Widget w, int height);
 static void setPaneMinHeight(Widget w, int min);
-static void updateWMSizeHints(WindowInfo *window);
+static void addWindowIcon(Widget shell);
 static void wmSizeUpdateProc(XtPointer clientData, XtIntervalId *id);
 #ifdef ROWCOLPATCH
 static void patchRowCol(Widget w);
@@ -91,15 +97,17 @@ static void patchedRemoveChild(Widget child);
 /*
 ** Create a new editor window
 */
-WindowInfo *CreateWindow(char *name, int rows, int cols, int autoIndent, 
-	int autoSave, int saveOldVersion, char *fontName, XmFontList fontList,
-	int wrap, int overstrike, int tabDist, int showMatching)
+WindowInfo *CreateWindow(char *name)
 {
     Widget appShell, main, menuBar, pane, text, stats;
     WindowInfo *window;
     Arg al[20];
     int ac;
-    static Pixmap iconPixmap = 0, maskPixmap = 0;
+#ifdef SGI_CUSTOM
+    char title_string[MAXPATHLEN + 14 + SGI_WINDOW_TITLE_LEN] = SGI_WINDOW_TITLE; 
+#else
+    char title_string[MAXPATHLEN + 2 + MAXPATHLEN + 14];
+#endif
 
     /* Allocate some memory for the new window data structure */
     window = (WindowInfo *)XtMalloc(sizeof(WindowInfo));
@@ -129,31 +137,52 @@ WindowInfo *CreateWindow(char *name, int rows, int cols, int autoIndent,
     window->undoMemUsed = 0;
     window->readOnly = FALSE;
     window->lockWrite = FALSE;
-    window->autoIndent = autoIndent;
-    window->autoSave = autoSave;
-    window->saveOldVersion = saveOldVersion;
-    window->wrap = wrap;
-    window->overstrike = overstrike;
-    window->showMatching = showMatching;
+    window->indentStyle = GetPrefAutoIndent(PLAIN_LANGUAGE_MODE);
+    window->autoSave = GetPrefAutoSave();
+    window->saveOldVersion = GetPrefSaveOldVersion();
+    window->wrapMode = GetPrefWrap(PLAIN_LANGUAGE_MODE);
+    window->overstrike = False;
+    window->showMatching = GetPrefShowMatching();
     window->showStats = GetPrefStatsLine();
+    window->highlightSyntax = GetPrefHighlightSyntax();
     window->modeMessageDisplayed = FALSE;
     window->ignoreModify = FALSE;
     window->windowMenuValid = FALSE;
     window->prevOpenMenuValid = FALSE;
     window->flashTimeoutID = 0;
     window->wasSelected = FALSE;
-    strcpy(window->fontName, fontName);
-    window->fontList = fontList;
-    window->shellCmdDialog = NULL;
-    window->macroCmdDialog = NULL;
+    strcpy(window->fontName, GetPrefFontName());
+    strcpy(window->italicFontName, GetPrefItalicFontName());
+    strcpy(window->boldFontName, GetPrefBoldFontName());
+    strcpy(window->boldItalicFontName, GetPrefBoldItalicFontName());
+    window->fontList = GetPrefFontList();
+    window->italicFontStruct = GetPrefItalicFont();
+    window->boldFontStruct = GetPrefBoldFont();
+    window->boldItalicFontStruct = GetPrefBoldItalicFont();
+    window->fontDialog = NULL;
     window->nMarks = 0;
     window->markTimeoutID = 0;
+    window->highlightData = NULL;
+    window->shellCmdData = NULL;
+    window->macroCmdData = NULL;
+    window->smartIndentData = NULL;
+    window->languageMode = PLAIN_LANGUAGE_MODE;
     
     /* Create a new toplevel shell to hold the window */
     ac = 0;
-    XtSetArg(al[ac], XmNtitle, name); ac++;
+#ifndef SGI_CUSTOM
+    strcpy(title_string, GetPrefServerName());
+    if (title_string[0] != '\0')
+        strcat (title_string, ": ");
+#endif
+    strcat(title_string, name);
+    XtSetArg(al[ac], XmNtitle, title_string); ac++;
     XtSetArg(al[ac], XmNdeleteResponse, XmDO_NOTHING); ac++;
-    XtSetArg(al[ac], XmNiconName, name); ac++;
+    if (strncmp(name, "Untitled", 8) == 0) { 
+    	XtSetArg(al[ac], XmNiconName, APP_NAME); ac++;
+    } else {
+    	XtSetArg(al[ac], XmNiconName, name); ac++;
+    }
     appShell = XtAppCreateShell("nedit", "NEdit",
 		applicationShellWidgetClass, TheDisplay, al, ac);
     window->shell = appShell;
@@ -163,19 +192,9 @@ WindowInfo *CreateWindow(char *name, int rows, int cols, int autoIndent,
 	    _XEditResCheckMessages, NULL);
 #endif /* EDITRES */
 
-    /* Add the icon.  There is something weird about the  XmNiconPixmap
-       resource that prevents it from being set from the defaults in
-       the application resource database.  It's just hardwired here. */
-    if (iconPixmap == 0) {
-    	iconPixmap = XCreateBitmapFromData(TheDisplay,
-    		RootWindowOfScreen(XtScreen (appShell)), (char *)iconBits,
-    		iconBitmapWidth, iconBitmapHeight);
-    	maskPixmap = XCreateBitmapFromData(TheDisplay,
-    		RootWindowOfScreen(XtScreen (appShell)), (char *)maskBits,
-    		iconBitmapWidth, iconBitmapHeight);
-    }
-    XtVaSetValues(appShell, XmNiconPixmap, iconPixmap,
-    	    XmNiconMask, maskPixmap, 0);
+#ifndef SGI_CUSTOM
+    addWindowIcon(appShell);
+#endif
 
     /* Create a MainWindow to manage the menubar and text area, set the
        userData resource to be used by WidgetToWindow to recover the
@@ -220,28 +239,36 @@ WindowInfo *CreateWindow(char *name, int rows, int cols, int autoIndent,
 
     /* Create the first, and most permanent text area (other panes may
        be added & removed, but this one will never be removed */
-    text = createTextArea(pane, window, rows, cols);
+    text = createTextArea(pane, window, GetPrefRows(), GetPrefCols(),
+    	    GetPrefEmTabDist(PLAIN_LANGUAGE_MODE), GetPrefDelimiters(),
+	    GetPrefWrapMargin());
     XtManageChild(text);
     window->textArea = text;
     window->lastFocus = text;
 
-#ifdef RIGHT_POPUP
     /* Create the right button popup menu (note: order is important here,
-       since the event handler for popping up this menu was probably already
-       added in createTextArea, but CreatePopupMenu requires window->textArea
+       since the translation for popping up this menu was probably already
+       added in createTextArea, but CreateBGMenu requires window->textArea
        to be set so it can attach the menu to it (because menu shells are
        finicky about the kinds of widgets they are attached to)) */
-    window->popMenu = CreatePopupMenu(window);
-#endif    
+    window->bgMenuPane = CreateBGMenu(window);
     
-    /* Use this permanent text widget's buffer as the buffer to hold all text
-       in the editor and the widget as the owner for buffer selections */
-    window->buffer = TextGetBuffer(text);
+    /* Create the text buffer rather than using the one created automatically
+       with the text area widget.  This is done so the syntax highlighting
+       modify callback can be called to synchronize the style buffer BEFORE
+       the text display's callback is called upon to display a modification */
+    window->buffer = BufCreate();
+    BufAddModifyCB(window->buffer, SyntaxHighlightModifyCB, window);
+    
+    /* Attach the buffer to the text widget, and add callbacks for modify */
+    TextSetBuffer(text, window->buffer);
     BufAddModifyCB(window->buffer, modifiedCB, window);
+    
+    /* Designate the permanent text area as the owner for selections */
     HandleXSelections(text);
     
     /* Set the requested hardware tab distance and useTabs in the text buffer */
-    BufSetTabDistance(window->buffer, tabDist);
+    BufSetTabDistance(window->buffer, GetPrefTabDist(PLAIN_LANGUAGE_MODE));
     window->buffer->useTabs = GetPrefInsertTabs();
 
     /* add the window to the global window list, update the Windows menus */
@@ -255,7 +282,10 @@ WindowInfo *CreateWindow(char *name, int rows, int cols, int autoIndent,
     AddMotifCloseCallback(appShell, (XtCallbackProc)closeCB, window);
     
     /* Make window resizing work in nice character heights */
-    updateWMSizeHints(window);
+    UpdateWMSizeHints(window);
+    
+    /* Set the minimum pane height for the initial text pane */
+    UpdateMinPaneHeights(window);
     
     return window;
 }
@@ -265,17 +295,39 @@ WindowInfo *CreateWindow(char *name, int rows, int cols, int autoIndent,
 */
 void CloseWindow(WindowInfo *window)
 {
-    /* if this is the last window, don't close it, make it Untitled */
-    if (WindowList == window && window->next == NULL) {
+    int keepWindow;
+    char name[MAXPATHLEN];
+    
+    /* Free smart indent macro programs */
+    EndSmartIndent(window);
+    
+    /* Clean up macro references to the doomed window.  If a macro is
+       executing, stop it.  If macro is calling this (closing its own
+       window), leave the window alive until the macro completes */
+    keepWindow = !MacroWindowCloseActions(window);
+    
+#ifndef VMS
+    /* Kill shell sub-process and free related memory */
+    AbortShellCommand(window);
+#endif /*VMS*/
+    
+    /* if this is the last window, or must be kept alive temporarily because
+       it's running the macro calling us, don't close it, make it Untitled */
+    if (keepWindow || (WindowList == window && window->next == NULL)) {
+	window->filename[0] = '\0';
+	UniqueUntitledName(name);
 	window->readOnly = FALSE;
 	window->lockWrite = FALSE;
-	strcpy(window->filename, "Untitled");
+	window->fileMode = 0;
+	strcpy(window->filename, name);
 	strcpy(window->path, "");
 	window->ignoreModify = TRUE;
 	BufSetAll(window->buffer, "");
 	window->ignoreModify = FALSE;
 	window->filenameSet = FALSE;
 	window->fileChanged = FALSE;
+	StopHighlighting(window);
+	EndSmartIndent(window);
 	UpdateWindowTitle(window);
 	UpdateWindowReadOnly(window);
 	XtSetSensitive(window->closeItem, FALSE);
@@ -286,12 +338,17 @@ void CloseWindow(WindowInfo *window)
     	XmTextSetString(window->statsLine, ""); /* resets scroll pos of stats
     	    	    	    	    	           line from long file names */
     	UpdateStatsLine(window);
+	DetermineLanguageMode(window, True);
 	return;
     }
     
-    /* remove the buffer modification callback so the buffer will be
+    /* Free syntax highlighting patterns, if any. w/o redisplaying */
+    FreeHighlightingData(window);
+    	
+    /* remove the buffer modification callbacks so the buffer will be
        deallocated when the last text widget is destroyed */
     BufRemoveModifyCB(window->buffer, modifiedCB, window);
+    BufRemoveModifyCB(window->buffer, SyntaxHighlightModifyCB, window);
     
 #ifdef ROWCOLPATCH
     patchRowCol(window->menuBar);
@@ -336,7 +393,8 @@ void SplitWindow(WindowInfo *window)
     short paneHeights[MAX_PANES+1];
     int insertPositions[MAX_PANES+1], topLines[MAX_PANES+1];
     int horizOffsets[MAX_PANES+1];
-    int i, focusPane, totalHeight=0;
+    int i, focusPane, emTabDist, wrapMargin, totalHeight=0;
+    char *delimiters;
     Widget text;
     
     /* Don't create new panes if we're already at the limit */
@@ -359,13 +417,21 @@ void SplitWindow(WindowInfo *window)
     /* Unmanage & remanage the panedWindow so it recalculates pane heights */
     XtUnmanageChild(window->splitPane);
     
-    /* Create a text widget to add to the pane and set its buffer to be
-       the same as the other panes in the window */
-    text = createTextArea(window->splitPane, window, 1, 1);
+    /* Create a text widget to add to the pane and set its buffer and
+       highlight data to be the same as the other panes in the window */
+    XtVaGetValues(window->textArea, textNemulateTabs, &emTabDist,
+    	    textNwordDelimiters, &delimiters, textNwrapMargin, &wrapMargin, 0);
+    text = createTextArea(window->splitPane, window, 1, 1, emTabDist,
+    	    delimiters, wrapMargin);
     TextSetBuffer(text, window->buffer);
+    if (window->highlightData != NULL)
+    	AttachHighlightToWidget(text, window);
     XtManageChild(text);
     window->textPanes[window->nPanes++] = text;
     
+    /* Set the minimum pane height in the new pane */
+    UpdateMinPaneHeights(window);
+
     /* adjust the heights, scroll positions, etc., to split the focus pane */
     for (i=window->nPanes; i>focusPane; i--) {
     	insertPositions[i] = insertPositions[i-1];
@@ -496,7 +562,7 @@ void ShowStatsLine(WindowInfo *window, int state)
     }
     
     /* Tell WM that the non-expandable part of the window has changed size */
-    updateWMSizeHints(window);
+    UpdateWMSizeHints(window);
 }
 
 /*
@@ -524,7 +590,7 @@ void ClearModeMessage(WindowInfo *window)
 /*
 ** Count the windows
 */
-int NWindows()
+int NWindows(void)
 {
     WindowInfo *win;
     int n;
@@ -534,20 +600,32 @@ int NWindows()
 }
 
 /*
-** Turn on and off autoindent
+** Set autoindent state to one of  NO_AUTO_INDENT, AUTO_INDENT, or SMART_INDENT.
 */
 void SetAutoIndent(WindowInfo *window, int state)
 {
+    int autoIndent = state == AUTO_INDENT, smartIndent = state == SMART_INDENT;
     int i;
     
-    window->autoIndent = state;
-    XtVaSetValues(window->textArea, textNautoIndent, state, 0);
+    if (window->indentStyle == SMART_INDENT && !smartIndent)
+    	EndSmartIndent(window);
+    else if (smartIndent && window->indentStyle != SMART_INDENT)
+    	BeginSmartIndent(window, True);
+    window->indentStyle = state;
+    XtVaSetValues(window->textArea, textNautoIndent, autoIndent,
+    	    textNsmartIndent, smartIndent, 0);
     for (i=0; i<window->nPanes; i++)
-    	XtVaSetValues(window->textPanes[i], textNautoIndent, state, 0);
+    	XtVaSetValues(window->textPanes[i], textNautoIndent, autoIndent,
+    	    	textNsmartIndent, smartIndent, 0);
+    XmToggleButtonSetState(window->smartIndentItem, smartIndent, False);
+    XmToggleButtonSetState(window->autoIndentItem, autoIndent, False);
+    XmToggleButtonSetState(window->autoIndentOffItem, state == NO_AUTO_INDENT,
+    	    False);
 }
 
 /*
-** Set the font for a window from a font name (generates the fontList).
+** Set the fonts for "window" from a font name, and updates the display.
+** Also updates window->fontList which is used for statistics line.
 **
 ** Note that this leaks memory and server resources.  In previous NEdit
 ** versions, fontLists were carefully tracked and freed, but X and Motif
@@ -558,13 +636,91 @@ void SetAutoIndent(WindowInfo *window, int state)
 ** (which shouldn't happen much in normal NEdit operation), and skip the
 ** futile effort of freeing them.
 */
-void SetFontByName(WindowInfo *window, char *fontName)
+void SetFonts(WindowInfo *window, char *fontName, char *italicName,
+	char *boldName, char *boldItalicName)
 {
-    XFontStruct *font;
+    XFontStruct *font, *oldFont;
+    int i, oldFontWidth, oldFontHeight, fontWidth, fontHeight;
+    int borderWidth, borderHeight, marginWidth, marginHeight;
+    int primaryChanged, highlightChanged = False;
+    Dimension oldWindowWidth, oldWindowHeight, oldTextWidth, oldTextHeight;
+    Dimension textWidth, textHeight, newWindowWidth, newWindowHeight;
+    textDisp *textD = ((TextWidget)window->textArea)->text.textD;
     
-    font = XLoadQueryFont(TheDisplay, fontName);
-    recreateWindow(window, fontName, font==NULL ? NULL :
-    	    XmFontListCreate(font, XmSTRING_DEFAULT_CHARSET));
+    /* Check which fonts have changed */
+    primaryChanged = strcmp(fontName, window->fontName);
+    if (strcmp(italicName, window->italicFontName)) highlightChanged = True;
+    if (strcmp(boldName, window->boldFontName)) highlightChanged = True;
+    if (strcmp(boldItalicName, window->boldItalicFontName))
+    	highlightChanged = True;
+    if (!primaryChanged && !highlightChanged)
+    	return;
+    	
+    /* Get information about the current window sizing, to be used to
+       determine the correct window size after the font is changed */
+    XtVaGetValues(window->shell, XmNwidth, &oldWindowWidth, XmNheight,
+    	    &oldWindowHeight, 0);
+    XtVaGetValues(window->textArea, XmNheight, &textHeight, XmNwidth,
+    	    &textWidth, textNmarginHeight, &marginHeight, textNmarginWidth,
+    	    &marginWidth, textNfont, &oldFont, 0);
+    oldTextWidth = textWidth - 2*marginWidth;
+    oldTextHeight = textHeight - 2*marginHeight;
+    for (i=0; i<window->nPanes; i++) {
+    	XtVaGetValues(window->textPanes[i], XmNheight, &textHeight, 0);
+    	oldTextHeight += textHeight - 2*marginHeight;
+    }
+    borderWidth = oldWindowWidth - oldTextWidth;
+    borderHeight = oldWindowHeight - oldTextHeight;
+    oldFontWidth = oldFont->max_bounds.width;
+    oldFontHeight = textD->ascent + textD->descent;
+    
+    	
+    /* Change the fonts in the window data structure.  If the primary font
+       didn't work, use Motif's fallback mechanism by stealing it from the
+       statistics line.  Highlight fonts are allowed to be NULL, which
+       is interpreted as "use the primary font" */
+    if (primaryChanged) {
+	strcpy(window->fontName, fontName);
+	font = XLoadQueryFont(TheDisplay, fontName);
+	if (font == NULL)
+    	    XtVaGetValues(window->statsLine, XmNfontList, &window->fontList, 0);
+    	else
+	    window->fontList = XmFontListCreate(font, XmSTRING_DEFAULT_CHARSET);
+    }
+    if (highlightChanged) {
+	strcpy(window->italicFontName, italicName);
+	window->italicFontStruct = XLoadQueryFont(TheDisplay, italicName);
+	strcpy(window->boldFontName, boldName);
+	window->boldFontStruct = XLoadQueryFont(TheDisplay, boldName);
+	strcpy(window->boldItalicFontName, boldItalicName);
+	window->boldItalicFontStruct = XLoadQueryFont(TheDisplay, boldItalicName);
+    }
+
+    /* Change the primary font in all the widgets */
+    if (primaryChanged) {
+	font = GetDefaultFontStruct(window->fontList);
+	XtVaSetValues(window->textArea, textNfont, font, 0);
+	for (i=0; i<window->nPanes; i++)
+    	    XtVaSetValues(window->textPanes[i], textNfont, font, 0);
+    }
+    
+    /* Change the highlight fonts, even if they didn't change, because
+       primary font is read through the style table for syntax highlighting */
+    if (window->highlightData != NULL)
+    	UpdateHighlightStyles(window);
+        
+    /* Use the information from the old window to re-size the window to a
+       size appropriate for the new font */
+    fontWidth = GetDefaultFontStruct(window->fontList)->max_bounds.width;
+    fontHeight = textD->ascent + textD->descent;
+    newWindowWidth = (oldTextWidth*fontWidth) / oldFontWidth + borderWidth;
+    newWindowHeight = (oldTextHeight*fontHeight) / oldFontHeight + borderHeight;
+    XtVaSetValues(window->shell, XmNwidth, newWindowWidth, XmNheight,
+    	    newWindowHeight, 0);
+    
+    /* Change the window manager size hints and the minimum pane height */
+    UpdateWMSizeHints(window);
+    UpdateMinPaneHeights(window);
 }
 
 /*
@@ -581,16 +737,23 @@ void SetOverstrike(WindowInfo *window, int overstrike)
 }
 
 /*
-** Turn on/off auto-wrap mode
+** Select auto-wrap mode, one of NO_WRAP, NEWLINE_WRAP, or CONTINUOUS_WRAP
 */
 void SetAutoWrap(WindowInfo *window, int state)
 {
     int i;
+    int autoWrap = state == NEWLINE_WRAP, contWrap = state == CONTINUOUS_WRAP;
 
-    XtVaSetValues(window->textArea, textNautoWrap, state, 0);
+    XtVaSetValues(window->textArea, textNautoWrap, autoWrap,
+    	    textNcontinuousWrap, contWrap, 0);
     for (i=0; i<window->nPanes; i++)
-    	XtVaSetValues(window->textPanes[i], textNautoWrap, state, 0);
-    window->wrap = state;
+    	XtVaSetValues(window->textPanes[i], textNautoWrap, autoWrap,
+    	    	textNcontinuousWrap, contWrap, 0);
+    window->wrapMode = state;
+    
+    XmToggleButtonSetState(window->newlineWrapItem, autoWrap, False);
+    XmToggleButtonSetState(window->continuousWrapItem, contWrap, False);
+    XmToggleButtonSetState(window->noWrapItem, state == NO_WRAP, False);
 }
 
 /*
@@ -627,65 +790,6 @@ WindowInfo *WidgetToWindow(Widget w)
     return window;
 }
 
-/*
-** To reset font, it is necessary to close and re-create the whole window.
-** (the paned window won't allow width resize required for font change).
-*/
-static void recreateWindow(WindowInfo *window, char *fontName,
-	XmFontList fontList)
-{
-    WindowInfo *newWindow;
-    int i, rows, cols, topLine, horizOffset;
-    int emTabDist, wrapMargin;
-    char *s1;
-    
-    /* create a new window and transfer values from the old one */
-    XtVaGetValues(window->textArea, textNrows, &rows, textNcolumns, &cols,
-    	    textNemulateTabs, &emTabDist, textNwrapMargin, &wrapMargin, 0);
-    newWindow = CreateWindow("", rows, cols, window->autoIndent,
-    	    window->autoSave, window->saveOldVersion, fontName, fontList,
-    	    window->wrap, window->overstrike, window->buffer->tabDist,
-    	    window->showMatching);
-    XtVaSetValues(newWindow->textArea, textNemulateTabs, emTabDist,
-    	    textNwrapMargin, wrapMargin, 0);
-    XtSetSensitive(newWindow->undoItem, XtIsSensitive(window->undoItem));
-    newWindow->ignoreModify = TRUE;
-    BufSetAll(newWindow->buffer, s1=BufGetAll(window->buffer));
-    XtFree(s1);
-    newWindow->buffer->useTabs = window->buffer->useTabs;
-    newWindow->ignoreModify = FALSE;
-    TextSetCursorPos(newWindow->textArea, TextGetCursorPos(window->textArea));
-    TextGetScroll(window->textArea, &topLine, &horizOffset);
-    TextSetScroll(newWindow->textArea, topLine, horizOffset);
-    strcpy(newWindow->filename, window->filename);
-    strcpy(newWindow->path, window->path);
-    newWindow->fileMode = window->fileMode;
-    newWindow->undo = window->undo;
-    newWindow->redo = window->redo;
-    newWindow->autoSaveCharCount = window->autoSaveCharCount;
-    newWindow->autoSaveOpCount = window->autoSaveOpCount;
-    newWindow->undoOpCount = window->undoOpCount;
-    newWindow->undoMemUsed = window->undoMemUsed;
-    newWindow->filenameSet = window->filenameSet;
-    newWindow->fileChanged = window->fileChanged;
-    newWindow->readOnly = window->readOnly;
-    newWindow->lockWrite = window->lockWrite;
-    for (i=0; i<window->nMarks; i++)
-    	newWindow->markTable[i] = window->markTable[i];
-    newWindow->nMarks = window->nMarks;
-    UpdateWindowTitle(newWindow);
-    UpdateWindowReadOnly(newWindow);
-    
-    /* close the old window */
-#ifdef ROWCOLPATCH
-    patchRowCol(window->menuBar);
-#endif
-    BufRemoveModifyCB(window->buffer, modifiedCB, window);
-    XtDestroyWidget(window->shell);
-    removeFromWindowList(window);
-    XtFree((char *)window);
-}
-
 /*
 ** Change the window appearance and the window data structure to show
 ** that the file it contains has been modified
@@ -751,8 +855,19 @@ void UpdateWindowReadOnly(WindowInfo *window)
 
 void SetWindowTitle(WindowInfo *window, char *title)
 {
+#ifdef SGI_CUSTOM
+    char title_string[MAXPATHLEN + 14 + SGI_WINDOW_TITLE_LEN] = SGI_WINDOW_TITLE; 
+#else
+    char title_string[MAXPATHLEN + 2 + MAXPATHLEN + 14];
+
+    strcpy(title_string, GetPrefServerName());
+    if (title_string[0] != '\0')
+        strcat (title_string, ": ");
+#endif
+    strcat (title_string, title);
+
     /* Set both the window title and the icon title */
-    XtVaSetValues(window->shell, XmNtitle, title, XmNiconName, title, 0);
+    XtVaSetValues(window->shell, XmNtitle, title_string, XmNiconName, title, 0);
 
     /* Update the Windows menus with the new name */
     InvalidateWindowMenus();
@@ -887,13 +1002,10 @@ void MakeSelectionVisible(WindowInfo *window, Widget textPane)
 }
 
 static Widget createTextArea(Widget parent, WindowInfo *window, int rows,
-	int cols)
+	int cols, int emTabDist, char *delimiters, int wrapMargin)
 {
     Widget text, sw, hScrollBar, vScrollBar;
     Pixel troughColor;
-    XFontStruct *fs;
-    int marginHeight;
-    Dimension hsbHeight, swMarginHeight;
         
     /* Create a text widget inside of a scrolled window widget */
     sw = XtVaCreateManagedWidget("scrolledW", xmScrolledWindowWidgetClass,
@@ -907,17 +1019,21 @@ static Widget createTextArea(Widget parent, WindowInfo *window, int rows,
     	    XmNrepeatDelay, 10, 0);
     text = XtVaCreateManagedWidget("text", textWidgetClass, sw,
     	    textNrows, rows, textNcolumns, cols,
-    	    textNemulateTabs, GetPrefEmTabDist(),
+    	    textNemulateTabs, emTabDist,
     	    textNfont, GetDefaultFontStruct(window->fontList),
     	    textNhScrollBar, hScrollBar, textNvScrollBar, vScrollBar,
-    	    textNwordDelimiters, GetPrefDelimiters(),
-    	    textNwrapMargin, GetPrefWrapMargin(),
-    	    textNautoIndent, window->autoIndent, textNautoWrap, window->wrap,
+	    textNreadOnly, window->readOnly || window->lockWrite,
+   	    textNwordDelimiters, delimiters,
+    	    textNwrapMargin, wrapMargin,
+    	    textNautoIndent, window->indentStyle == AUTO_INDENT,
+    	    textNsmartIndent, window->indentStyle == SMART_INDENT,
+    	    textNautoWrap, window->wrapMode == NEWLINE_WRAP,
+    	    textNcontinuousWrap, window->wrapMode == CONTINUOUS_WRAP,
     	    textNoverstrike, window->overstrike, 0);
     XtVaSetValues(sw, XmNworkWindow, text, XmNhorizontalScrollBar, hScrollBar,
     	    XmNverticalScrollBar, vScrollBar, 0);
     
-    /* add focus, drag, and cursor tracking callback */
+    /* add focus, drag, cursor tracking, and smart indent callbacks */
     XtAddCallback(text, textNfocusCallback, (XtCallbackProc)focusCB, window);
     XtAddCallback(text, textNcursorMovementCallback, (XtCallbackProc)movedCB,
     	    window);
@@ -925,7 +1041,8 @@ static Widget createTextArea(Widget parent, WindowInfo *window, int rows,
     	    window);
     XtAddCallback(text, textNdragEndCallback, (XtCallbackProc)dragEndCB,
     	    window);
-    
+    XtAddCallback(text, textNsmartIndentCallback, SmartIndentCB, window);
+    	    
     /* This makes sure the text area initially has a the insert point shown
        ... (check if still true with the nedit text widget, probably not) */
     XmAddTabGroup(XtParent(text));
@@ -934,22 +1051,13 @@ static Widget createTextArea(Widget parent, WindowInfo *window, int rows,
        bars to be the same color as the scroll bar interiors  */
     XtVaGetValues(vScrollBar, XmNtroughColor, &troughColor, 0);
     XtVaSetValues(sw, XmNbackground, troughColor, 0);
-
-    /* set the minimum size of a pane */
-    XtVaGetValues(sw, XmNscrolledWindowMarginHeight, &swMarginHeight, 0);
-    XtVaGetValues(text, textNmarginHeight, &marginHeight, textNfont, &fs, 0);
-    XtVaGetValues(hScrollBar, XmNheight, &hsbHeight, 0);
-    setPaneMinHeight(sw, fs->ascent + fs->descent + marginHeight*2 +
-    	    swMarginHeight*2 + hsbHeight);
     
     /* compensate for Motif delete/backspace problem */
     RemapDeleteKey(text);
 
-#ifdef RIGHT_POPUP
-    /* Attach event handler for right button popup menu */
-    AddPopupHandler(window, text);
-#endif    
-    
+    /* Augment translation table for right button popup menu */
+    AddBGMenuAction(text);
+   
     return text;
 }
 
@@ -983,10 +1091,8 @@ static void modifiedCB(int pos, int nInserted, int nDeleted, int nRestyled,
 #ifndef VMS
     	XtSetSensitive(window->filterItem, selected);
 #endif
-#ifdef RIGHT_POPUP
-    	XtSetSensitive(window->popCutItem, selected);
-    	XtSetSensitive(window->popCopyItem, selected);
-#endif
+
+	DimSelectionDepUserMenuItems(window, selected);
     	if (window->replaceDlog != NULL)
     	    XtSetSensitive(window->replaceInSelBtn, selected);
     }
@@ -1049,7 +1155,9 @@ static void dragEndCB(Widget w, WindowInfo *window, dragEndCBStruct *callData)
 static void closeCB(Widget w, WindowInfo *window, XtPointer callData) 
 {
     if (WindowList->next == NULL) {
-    	if (!WindowList->fileChanged)
+	if (!CheckPrefsChangesSaved(window->shell))
+    	    return;
+	if (!WindowList->fileChanged)
      	    exit(0);
      	if (CloseFileAndWindow(window))
      	    exit(0);
@@ -1094,9 +1202,12 @@ static void removeFromWindowList(WindowInfo *window)
 */
 void UpdateStatsLine(WindowInfo *window)
 {
-    int line, length, pos, colNum, topLineNum, horizOffset;
+    int line, pos, colNum;
     char *string;
     Widget statW = window->statsLine;
+#ifdef SGI_CUSTOM
+    char *sleft, *smid, *sright;
+#endif
     
     /* This routine is called for each character typed, so its performance
        affects overall editor perfomance.  Only update if the line is on
@@ -1106,20 +1217,31 @@ void UpdateStatsLine(WindowInfo *window)
     
     /* Compose the string to display. If line # isn't available, leave it off */
     pos = TextGetCursorPos(window->lastFocus);
-    length = window->buffer->length;
     string = XtMalloc(strlen(window->filename) + strlen(window->path) + 45);
-    if (!TextPosToLineNum(window->lastFocus, pos, &line))
-    	sprintf(string, "%s%s %d bytes", window->path, window->filename,length);
-    else {
-    	TextGetScroll(window->lastFocus, &topLineNum, &horizOffset);
-    	colNum = BufCountDispChars(window->buffer,
-    		BufStartOfLine(window->buffer, pos), pos);
-    	sprintf(string, "%s%s line %d, col %d, %d bytes",
-    		window->path, window->filename, line, colNum, length);
-    }
+    if (!TextPosToLineAndCol(window->lastFocus, pos, &line, &colNum))
+    	sprintf(string, "%s%s %d bytes", window->path, window->filename,
+    	    	window->buffer->length);
+    else
+    	sprintf(string, "%s%s line %d, col %d, %d bytes", window->path,
+    	    	window->filename, line, colNum, window->buffer->length);
     
     /* Change the text in the stats line */
+#ifdef SGI_CUSTOM
+    /* don't show full pathname, just dir and filename (+ line/col/byte info) */
+    smid = strchr(string, '/'); 
+    if ( smid != NULL ) {
+	sleft = smid;
+	sright = strrchr(string, '/'); 
+        while (strcmp(smid, sright)) {
+		sleft = smid;
+	  	smid = strchr(sleft + 1, '/');
+	}
+    	XmTextReplace(statW, 0, XmTextGetLastPosition(statW), sleft + 1);
+    } else
+    	XmTextReplace(statW, 0, XmTextGetLastPosition(statW), string);
+#else
     XmTextReplace(statW, 0, XmTextGetLastPosition(statW), string);
+#endif
     XtFree(string);
 }
 
@@ -1147,12 +1269,14 @@ static void setPaneMinHeight(Widget w, int min)
 ** hints also makes the resize indicator show the window size in characters
 ** rather than pixels, which is very helpful to users.
 */
-static void updateWMSizeHints(WindowInfo *window)
+void UpdateWMSizeHints(WindowInfo *window)
 {
-    Dimension shellWidth, shellHeight, textHeight, textWidth;
+    Dimension shellWidth, shellHeight, textHeight, textWidth, hScrollBarHeight;
     int marginHeight, marginWidth, totalHeight;
     XFontStruct *fs;
     int i, baseWidth, baseHeight, fontHeight, fontWidth;
+    Widget hScrollBar;
+    textDisp *textD = ((TextWidget)window->textArea)->text.textD;
 
     /* Find the base (non-expandable) width and height of the editor window */
     XtVaGetValues(window->textArea, XmNheight, &textHeight, XmNwidth,
@@ -1160,8 +1284,13 @@ static void updateWMSizeHints(WindowInfo *window)
     	    &marginWidth, 0);
     totalHeight = textHeight - 2*marginHeight;
     for (i=0; i<window->nPanes; i++) {
-    	XtVaGetValues(window->textPanes[i], XmNheight, &textHeight, 0);
+    	XtVaGetValues(window->textPanes[i], XmNheight, &textHeight, 
+    		textNhScrollBar, &hScrollBar, 0);
     	totalHeight += textHeight - 2*marginHeight;
+    	if (!XtIsManaged(hScrollBar)) {
+    	    XtVaGetValues(hScrollBar, XmNheight, &hScrollBarHeight, 0);
+    	    totalHeight -= hScrollBarHeight;
+    	}
     }
     XtVaGetValues(window->shell, XmNwidth, &shellWidth,
     	    XmNheight, &shellHeight, 0);
@@ -1170,7 +1299,7 @@ static void updateWMSizeHints(WindowInfo *window)
     
     /* Find the dimensions of a single character of the text font */
     XtVaGetValues(window->textArea, textNfont, &fs, 0);
-    fontHeight = fs->ascent + fs->descent;
+    fontHeight = textD->ascent + textD->descent;
     fontWidth = fs->max_bounds.width;
     
     /* Set the size hints in the shell widget */
@@ -1181,6 +1310,69 @@ static void updateWMSizeHints(WindowInfo *window)
     	    XmNminHeight, baseHeight + (1+window->nPanes) * fontHeight, 0);
 }
 
+/*
+** Update the minimum allowable height for a split window pane after a change
+** to font or margin height.
+*/
+void UpdateMinPaneHeights(WindowInfo *window)
+{
+    textDisp *textD = ((TextWidget)window->textArea)->text.textD;
+    Dimension hsbHeight, swMarginHeight;
+    int i, marginHeight, minPaneHeight;
+    Widget hScrollBar;
+
+    /* find the minimum allowable size for a pane */
+    XtVaGetValues(window->textArea, textNhScrollBar, &hScrollBar, 0);
+    XtVaGetValues(XtParent(window->textArea), XmNscrolledWindowMarginHeight,
+    	    &swMarginHeight, 0);
+    XtVaGetValues(window->textArea, textNmarginHeight, &marginHeight, 0);
+    XtVaGetValues(hScrollBar, XmNheight, &hsbHeight, 0);
+    minPaneHeight = textD->ascent + textD->descent + marginHeight*2 +
+    	    swMarginHeight*2 + hsbHeight;
+    
+    /* Set it in all of the widgets in the window */
+    setPaneMinHeight(XtParent(window->textArea), minPaneHeight);
+    for (i=0; i<window->nPanes; i++)
+    	setPaneMinHeight(XtParent(window->textPanes[i]), minPaneHeight);
+}
+
+/* Add an icon to an applicaction shell widget.  addWindowIcon adds a large
+** (primary window) icon, AddSmallIcon adds a small (secondary window) icon.
+**
+** Note: I would prefer that these were not hardwired, but yhere is something
+** weird about the  XmNiconPixmap resource that prevents it from being set
+** from the defaults in the application resource database.
+*/
+static void addWindowIcon(Widget shell)
+{ 
+    static Pixmap iconPixmap = 0, maskPixmap = 0;
+
+    if (iconPixmap == 0) {
+    	iconPixmap = XCreateBitmapFromData(TheDisplay,
+    		RootWindowOfScreen(XtScreen(shell)), (char *)iconBits,
+    		iconBitmapWidth, iconBitmapHeight);
+    	maskPixmap = XCreateBitmapFromData(TheDisplay,
+    		RootWindowOfScreen(XtScreen(shell)), (char *)maskBits,
+    		iconBitmapWidth, iconBitmapHeight);
+    }
+    XtVaSetValues(shell, XmNiconPixmap, iconPixmap, XmNiconMask, maskPixmap, 0);
+}
+void AddSmallIcon(Widget shell)
+{ 
+    static Pixmap iconPixmap = 0, maskPixmap = 0;
+
+    if (iconPixmap == 0) {
+    	iconPixmap = XCreateBitmapFromData(TheDisplay,
+    		RootWindowOfScreen(XtScreen(shell)), (char *)n_bits,
+    		n_width, n_height);
+    	maskPixmap = XCreateBitmapFromData(TheDisplay,
+    		RootWindowOfScreen(XtScreen(shell)), (char *)n_mask,
+    		n_width, n_height);
+    }
+    XtVaSetValues(shell, XmNiconPixmap, iconPixmap,
+    	    XmNiconMask, maskPixmap, 0);
+}
+
 /*
 ** Xt timer procedure for updating size hints.  The new sizes of objects in
 ** the window are not ready immediately after adding or removing panes.  This
@@ -1190,7 +1382,7 @@ static void updateWMSizeHints(WindowInfo *window)
 */
 static void wmSizeUpdateProc(XtPointer clientData, XtIntervalId *id)
 {
-    updateWMSizeHints((WindowInfo *)clientData);
+    UpdateWMSizeHints((WindowInfo *)clientData);
 }
 
 #ifdef ROWCOLPATCH
